/*
    This file is part of Structorizer.

    Structorizer is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Structorizer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

 ***********************************************************************

    OBERON Source Code Generator

    Copyright (C) 2008 Klaus-Peter Reimers

    This file has been released under the terms of the GNU General
    Public License as published by the Free Software Foundation.

 */

package lu.fisch.structorizer.generators;

/******************************************************************************************************
 *
 *      Author:         Klaus-Peter Reimers
 *
 *      Description:    This class generates Oberon code.
 *
 ******************************************************************************************************
 *
 *      Revision List
 *
 *      Author					Date			Description
 *      ------                      		----			-----------
 *      Klaus-Peter Reimers                     2008.01.08              First Issue
 *	Bob Fisch				2008.01.08		Modified "private String transform(String _input)"
 *	Bob Fisch				2008.04.12		Added "Fields" section for generator to be used as plugin
 *	Bob Fisch				2008.08.14		Added declaration output. A comment line in the root element
 *												with a "#" is ignored. All other lines are written to the code.
 *
 ******************************************************************************************************
 *
 *      Comment:		Based on "PasGenerator.java" from Bob Fisch
 *
 ******************************************************************************************************///

import lu.fisch.utils.*;
import lu.fisch.structorizer.parsers.*;
import lu.fisch.structorizer.elements.*;

public class OberonGenerator extends Generator {
	
	/************ Fields ***********************/
	protected String getDialogTitle()
	{
		return "Export Oberon Code ...";
	}
	
	protected String getFileDescription()
	{
		return "Oberon Source Code";
	}
	
	protected String getIndent()
	{
		return "  ";
	}
	
	protected String[] getFileExtensions()
	{
		String[] exts = {"Mod"};
		return exts;
	}
	
	/************ Code Generation **************/
	private String transform(String _input)
	{
		// et => and
		// ou => or
		// lire => readln()
		// écrire => writeln()
		// tant que => ""
		// pour => ""
		// jusqu'à => ""
		// à => "to"
	
		_input=BString.replace(_input," <- ",":=");
		_input=BString.replace(_input,"<- ",":=");
		_input=BString.replace(_input," <-",":=");
		_input=BString.replace(_input,"<-",":=");
		
		// BEGIN: Added 2008.01.07 by Bob Fisch
		if(!D7Parser.preAlt.equals("")){_input=BString.replace(_input,D7Parser.preAlt,"");}
		if(!D7Parser.postAlt.equals("")){_input=BString.replace(_input,D7Parser.postAlt,"");}
		if(!D7Parser.preCase.equals("")){_input=BString.replace(_input,D7Parser.preCase,"");}
		if(!D7Parser.postCase.equals("")){_input=BString.replace(_input,D7Parser.postCase,"");}
		if(!D7Parser.preFor.equals("")){_input=BString.replace(_input,D7Parser.preFor,"");}
		if(!D7Parser.postFor.equals("")){_input=BString.replace(_input,D7Parser.postFor,"TO");}
		if(!D7Parser.preWhile.equals("")){_input=BString.replace(_input,D7Parser.preWhile,"");}
		if(!D7Parser.postWhile.equals("")){_input=BString.replace(_input,D7Parser.postWhile,"");}
		if(!D7Parser.preRepeat.equals("")){_input=BString.replace(_input,D7Parser.preRepeat,"");}
		if(!D7Parser.postRepeat.equals("")){_input=BString.replace(_input,D7Parser.postRepeat,"");}
		// END: Added 2008.01.07 by Bob Fisch

		return _input.trim();
	}
	
	protected void generateCode(Instruction _inst, String _indent)
	{
		for(int i=0;i<_inst.getText().count();i++)
		{
			code.add(_indent+transform(_inst.getText().get(i))+";");
		}
	}
	
	protected void generateCode(Alternative _alt, String _indent)
	{
		code.add(_indent+"IF "+BString.replace(transform(_alt.getText().getText()),"\n","")+" THEN");
		generateCode(_alt.qTrue,_indent+_indent);
		if(_alt.qFalse.getSize()!=0)
		{
			code.add(_indent+"END");
			code.add(_indent+"ELSE");
			generateCode(_alt.qFalse,_indent+_indent);
		}
		code.add(_indent+"END;");
	}
	
	protected void generateCode(Case _case, String _indent)
	{
		code.add(_indent+"CASE "+transform(_case.getText().get(0))+" OF");
		
		for(int i=0;i<_case.qs.size()-1;i++)
		{
			code.add(_indent+_indent.substring(0,1)+_case.getText().get(i+1).trim()+":");
			generateCode((Subqueue) _case.qs.get(i),_indent+_indent);
			code.add(_indent+"END;");
		}
		
		if(!_case.getText().get(_case.qs.size()).trim().equals("%"))
		{
			code.add(_indent+_indent.substring(0,1)+"ELSE");
			generateCode((Subqueue) _case.qs.get(_case.qs.size()-1),_indent+_indent.substring(0,1)+_indent.substring(0,1));
		}
		code.add(_indent+_indent.substring(0,1)+"END;");
	}
	
	protected void generateCode(For _for, String _indent)
	{
		code.add(_indent+"FOR "+BString.replace(transform(_for.getText().getText()),"\n","")+" DO");
		generateCode(_for.q,_indent+_indent);
		code.add(_indent+"END;");
	}
	
	protected void generateCode(While _while, String _indent)
	{
		code.add(_indent+"WHILE "+BString.replace(transform(_while.getText().getText()),"\n","")+" DO");
		generateCode(_while.q,_indent+_indent);
		code.add(_indent+"END;");
	}
	
	protected void generateCode(Repeat _repeat, String _indent)
	{
		code.add(_indent+"REPEAT");
		generateCode(_repeat.q,_indent+_indent);
		code.add(_indent+"UNTIL "+BString.replace(transform(_repeat.getText().getText()),"\n","")+";");
	}
	
	protected void generateCode(Forever _forever, String _indent)
	{
		code.add(_indent+"LOOP");
		generateCode(_forever.q,_indent+_indent);
		code.add(_indent+"END;");
	}
	
	protected void generateCode(Call _call, String _indent)
	{
		for(int i=0;i<_call.getText().count();i++)
		{
			code.add(_indent+transform(_call.getText().get(i))+";");
		}
	}
	
	protected void generateCode(Jump _jump, String _indent)
	{
		for(int i=0;i<_jump.getText().count();i++)
		{
			code.add(_indent+transform(_jump.getText().get(i))+";");
		}
	}
	
	protected void generateCode(Subqueue _subqueue, String _indent)
	{
		// code.add(_indent+"");
		for(int i=0;i<_subqueue.children.size();i++)
		{
			generateCode((Element) _subqueue.children.get(i),_indent);
		}
		// code.add(_indent+"");
	}
	
	public String generateCode(Root _root, String _indent)
	{
		String pr = "MODULE";
		String modname = _root.getText().get(0);
		if(_root.isProgram==false) {pr="PROCEDURE";}
		
		code.add(pr+" "+modname+";");
		code.add("");

		// Add comments and/or declarations to the program (Bob)
		for(int i=0;i<_root.getComment().count();i++)
		{
			if(!_root.getComment().get(i).startsWith("#"))
			{
				code.add(_root.getComment().get(i));
			}
		}
		
		//code.add("// declare your variables here");
		code.add("");
		code.add("BEGIN");
		generateCode(_root.children,_indent);
		code.add("END "+modname+".");
		
		return code.getText();
	}
	
	
}
